import torch as tc
from torch.utils.data import DataLoader

from trading_models.simple_models import CNN, MLP, Transformer
from trading_models.utils import WindowDataset, model_size

"""
trading models

input: x.shape = (T, F)
output: a.shape = (T-W+1, A)

T: time
F: features
W: window length
A: actions

at each time step t in range(W-1, T),
the model looks at data in the window
x[t+1-W : t+1, :]
to make A actions
"""

T, F, W, A = 100, 2, 50, 1
device = tc.device("cuda" if tc.cuda.is_available() else "cpu")
x = tc.randn(T, F).to(device)

dataset = WindowDataset(x, W)
dataloader = DataLoader(dataset, batch_size=32, shuffle=False)

net1 = MLP([W * F, 64, 64, A]).to(device)
net2 = CNN([F, 64, 64, A]).to(device)
net3 = Transformer(W, F, A, d_model=64, d_ff=64, n_head=2, n_layer=2).to(device)

for net in [net1, net2, net3]:
    outputs = []
    for batch in dataloader:
        outputs.append(net(batch.to(device)).detach())
    output = tc.cat(outputs, dim=0)
    print(net, output.shape, model_size(net), "\n")



from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sb
import torch as tc
import torch.nn as nn
import torch.nn.functional as F
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression, Ridge
from sklearn.metrics import mean_squared_error as mse
from sklearn.metrics import r2_score
from sklearn.neighbors import KNeighborsRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVR
from statsmodels.nonparametric.smoothers_lowess import lowess

from .utils import mlp, plot_general, tensor, to_np


def sanity_check(X: np.ndarray, Y: np.ndarray, models=None, use_lowess=False):
    F = X.shape[-1]
    # 1. basic plots
    plots = {}
    for i in range(F):
        x = X[:, i]
        plots[f"x{i}_hist"] = x
        plots[f"x{i}_vs_y"] = {"x": x, "y": Y}
        if use_lowess:
            r = lowess(Y, x).T
            plots[f"x{i}_vs_y_lowess"] = {"x": r[0], "y": r[1]}
    C = 3 if use_lowess else 2
    plots["y_hist"] = Y
    plot_general(plots, "ml_1_basic", C=C)

    # 2. correlation heatmap
    df = pd.DataFrame(X, columns=[f"x{i}" for i in range(F)])
    df["Y"] = Y
    plt.figure(figsize=(10, 10))
    sb.heatmap(df.corr(), annot=True)
    plt.savefig("ml_2_corr_heatmap")
    plt.close()

    # 3. PCA
    pca = PCA(n_components=2)
    x = pca.fit_transform(X).T
    plt.scatter(x[0], x[1], c=Y, s=2)
    plt.colorbar(label="Y")
    plt.xlabel("pca_x0")
    plt.ylabel("pca_x1")
    plt.title("PCA colored by Y")
    plt.savefig("ml_3_pca")
    plt.close()

    idx = int(len(X) * 0.6)
    X1, X2, Y1, Y2 = X[:idx], X[idx:], Y[:idx], Y[idx:]
    scaler = StandardScaler()
    X1 = scaler.fit_transform(X1)
    X2 = scaler.transform(X2)
    if models is None:
        models: List[SVR] = [
            LinearRegression(),
            Ridge(alpha=1.0),
            RandomForestRegressor(max_depth=6),
            SVR(epsilon=0.2),
            MLPRegressor(hidden_layer_sizes=[64, 64]),
            MyMLPRegressor(),
            KNeighborsRegressor(n_neighbors=10, weights="distance"),
        ]
    for m in models:
        m.fit(X1, Y1, X2, Y2) if isinstance(m, MyMLPRegressor) else m.fit(X1, Y1)
        name = m.__class__.__name__.replace("Regressor", "")
        parts = [
            f"{name:20} R^2: train {m.score(X1, Y1):.3f}, test {m.score(X2, Y2):.3f}",
            f"MSE: train {mse(Y1, m.predict(X1)):.1e}, test {mse(Y2, m.predict(X2)):.1e}",
        ]
        print("   ".join(parts))


class MyMLPRegressor:
    def __init__(s, hid_sizes=[64, 64], Act=nn.Tanh, lr=1e-3):
        s.hid_sizes, s.Act, s.lr = hid_sizes, Act, lr

    def fit(s, X: np.ndarray, Y: np.ndarray, X2: np.ndarray, Y2: np.ndarray):
        min_steps = 100
        s.net = mlp([X.shape[-1], *s.hid_sizes, 1], s.Act)
        opt = tc.optim.Adam(s.net.parameters(), lr=s.lr)
        X, Y = tensor([X, Y.reshape(-1, 1)])
        X2, Y2 = tensor([X2, Y2.reshape(-1, 1)])
        losses, test_losses = [], []
        while True:
            with tc.no_grad():
                test_loss = F.mse_loss(s.net(X2), Y2).item()
                test_losses.append(test_loss)
            opt.zero_grad()
            loss = F.mse_loss(s.net(X), Y)
            loss.backward()
            opt.step()

            loss = loss.item()
            losses.append(loss)
            # print(f"MyMLP mse_loss: train {loss:.1e}, test {test_loss:.1e}")
            avg = np.mean(losses[-min_steps:])
            if len(losses) > min_steps and loss > 0.95 * avg:
                break
        plot_general({"loss": losses, "test_loss": test_losses}, "MyMLP_loss")

    @tc.no_grad()
    def predict(s, X):
        return to_np(s.net(tensor(X))).squeeze()

    def score(s, X: np.ndarray, Y: np.ndarray):
        return r2_score(Y, s.predict(X))


def checkerboard_data(N=int(3e4), C=10, plot=True, noise=0.5):
    X = np.random.uniform(-1, 1, (N, 2))
    F = np.floor
    mask = (F(C * X[:, 0]) + F(C * X[:, 1])) % 2 == 0
    Y = np.where(mask, 1, -1) + np.random.normal(0, noise, N)
    if plot:
        plt.scatter(X[:, 0], X[:, 1], c=Y, s=3)
        plt.show()
    return X, Y



import time
from typing import Any, Callable, Dict, List, Tuple

import matplotlib.pyplot as plt
import numpy as np
import torch as tc
import torch.nn as nn
from PIL import Image
from torch.utils.data import Dataset

DEVICE = tc.device("cuda" if tc.cuda.is_available() else "cpu")


def std_clip(x: np.ndarray, n=3):
    mean, std = x.mean(), x.std() * n
    return x.clip(mean - std, mean + std)


def mlp(sizes, Act=nn.Tanh, out=[]):
    layers = []
    for a, b in zip(sizes[:-1], sizes[1:]):
        layers += [nn.Linear(a, b), Act()]
    return nn.Sequential(*layers[:-1], *out)


def model_size(m: nn.Module):
    a = sum(p.numel() for p in m.parameters())
    b = sum(p.numel() for p in m.parameters() if p.requires_grad)
    return f"trainable: {b}/{a}"


class WindowDataset(Dataset):
    def __init__(s, x, W):
        s.x, s.W = x, W

    def __len__(s):
        return len(s.x) - s.W + 1

    def __getitem__(s, idx):
        return s.x[idx : idx + s.W]


# ==============================

D_TYPE = Dict[str, np.ndarray]
D2_TYPE = Dict[str, D_TYPE]
T2_TYPE = Dict[str, Dict[str, tc.Tensor]]
LOSS_FN_TYPE = Callable[[T2_TYPE, str], Tuple[tc.Tensor, Dict, Dict]]


def train_model(
    net: nn.Module,
    train_data: Any,
    test_data: Any,
    loss_fn: LOSS_FN_TYPE,
    id="train_model",
    lr=1e-3,
    n_epoch=10000,
    f_test=10,
    f_plot=20,
):
    net.to(DEVICE)
    opt = tc.optim.AdamW(net.parameters(), lr=lr)
    records, gif = [], GIFMaker()
    best_score = -np.inf

    for e in range(int(n_epoch)):
        loss_test, info_test = np.nan, {}
        if e % f_test == 0:
            net.eval()
            with tc.no_grad():
                loss_test, info_test, plots_test = loss_fn(test_data, "test")
            score = info_test["SCORE_test"]
            if score > best_score:
                best_score = score
                tc.save(net.state_dict(), f"{id}.tc")
            net.train()
        opt.zero_grad()
        loss_train, info_train, plots_train = loss_fn(train_data, "train")
        loss_train.backward()
        opt.step()
        rec = {"A_loss_train": loss_train, "A_loss_test": loss_test}
        records.append({**rec, **info_train, **info_test})
        if e % f_plot == 0:
            plots = transpose_records(records)
            plots = {**plots, **plots_train, **plots_test}
            plot_general(dict(sorted(plots.items())), f"{id}.png")
            gif.add(f"{id}.png")
    gif.save(id)


def mod_keys(dic: Dict, post):
    return {f"{k}_{post}": v for k, v in dic.items()}


def slice_xy(xy: D2_TYPE, r1, r2, num=None):
    xy2 = {}
    for sym, d in xy.items():
        x, y = d["x"], d["y"]
        n = len(x)
        i1, i2 = int(r1 * n), int(r2 * n)
        step = max(1, (i2 - i1) // num) if num else 1
        x, y = x[i1:i2:step], y[i1:i2:step]
        xy2[sym] = dict(x=x, y=y)
    print(f"slice_xy: n: {n}, i1: {i1}, i2: {i2}, step: {step}, n2: {len(x)}")
    return xy2


def concat_xy(xy: D2_TYPE):
    X = np.concat([d["x"] for d in xy.values()], axis=0)
    Y = np.concat([d["y"] for d in xy.values()], axis=0)
    return {"ALL_SYMBOLS": dict(x=X, y=Y)}


def timer(func):
    def func2(*args, **kwargs):
        t1 = time.time()
        res = func(*args, **kwargs)
        print(f"{func.__name__} took {time.time() - t1:.3f} seconds")
        return res

    return func2


def tensor(x):
    if isinstance(x, np.ndarray):
        return tc.from_numpy(x.copy()).float().to(DEVICE)
    if isinstance(x, dict):
        return {k: tensor(v) for k, v in x.items()}
    if isinstance(x, (list, tuple)):
        return [tensor(v) for v in x]
    return x


def to_np(x):
    if isinstance(x, tc.Tensor):
        return x.detach().cpu().numpy()
    if isinstance(x, dict):
        return {k: to_np(v) for k, v in x.items()}
    if isinstance(x, (list, tuple)):
        return [to_np(v) for v in x]
    return x


def shape(x):
    if isinstance(x, (np.ndarray, tc.Tensor)):
        return x.shape
    if isinstance(x, dict):
        return {k: shape(v) for k, v in x.items()}
    if isinstance(x, (list, tuple)):
        return [shape(v) for v in x]
    return x


def slice_data(x, r1, r2, axis):
    if isinstance(x, (np.ndarray, tc.Tensor)):
        N = x.shape[axis]
        slc = [slice(None)] * x.ndim
        slc[axis] = slice(int(N * r1), int(N * r2))
        return x[tuple(slc)]
    if isinstance(x, dict):
        return {k: slice_data(v, r1, r2, axis) for k, v in x.items()}
    if isinstance(x, (list, tuple)):
        return [slice_data(v, r1, r2, axis) for v in x]
    return x


def l2_norm(x: nn.Module):
    return sum(p.norm(2) for p in x.parameters())


def add_l2_loss(loss: tc.Tensor, net: nn.Module, ratio):
    l2 = l2_norm(net)
    loss += l2 / l2.item() * loss.abs().item() * ratio
    return loss, l2


# ==================================


def plot_general(plots: D_TYPE, id, C=2, bins=100):
    R = int(np.ceil(len(plots) / C))
    plt.figure(figsize=(4 * C, 3 * R))
    i = 0
    for k, v in plots.items():
        v = to_np(v)
        i += 1
        plt.subplot(R, C, i)
        title = k
        if "_hist" in k:
            v = v.flatten()
            if len(v):
                mean, std = v.mean(), v.std() * 3
                plt.hist(v, bins=bins, range=[mean - std, mean + std])
                title = f"{k}\n N={len(v):.1e}, {v.min():.5g} ~ {v.max():.5g}"
        elif isinstance(v, dict):
            plt.scatter(v["x"], v["y"], s=3)
            corr = np.corrcoef(v["x"], v["y"])[0, 1]
            title = f"{k}\n corr={corr:.3g}"
        else:
            x = np.arange(len(v))
            if np.any(np.isnan(v)):
                plt.scatter(x, v, s=3)
            else:
                plt.plot(x, v)
        plt.title(title)
    plt.tight_layout()
    plt.savefig(id)
    plt.close()


def transpose_records(records: List[Dict]):
    return {k: np.array([to_np(r.get(k, np.nan)) for r in records]) for k in records[0]}


def plot_records(records: List[Dict], id, C=1):
    dic = transpose_records(records)
    R = int(np.ceil(len(dic) / C))
    plt.figure(figsize=(4 * C, 3 * R))
    i = 0
    for k, v in dic.items():
        i += 1
        plt.subplot(R, C, i)
        plt.title(k)
        x = np.arange(len(v))
        if np.any(np.isnan(v)):
            plt.scatter(x, v, s=3)
        else:
            plt.plot(x, v)
    plt.tight_layout()
    plt.savefig(id)
    plt.close()


def plot_xy(xy: D2_TYPE, id="xy"):
    R, C = 5, 2
    plt.figure(figsize=(4 * C, 3 * R))
    i = 0
    for sym, d in xy.items():
        i += 1
        x, y = d["x"], d["y"]
        plt.subplot(R, C, i)
        plt.title(sym)
        for k in range(x.shape[1]):
            plt.plot(x[:, k], label=f"x[:, {k}]")
        plt.legend()
        i += 1
        plt.subplot(R, C, i)
        plt.plot(y)
        if i == R * C:
            break
    plt.tight_layout()
    plt.savefig(id)
    plt.close()


# =====================


class GIFMaker:
    def __init__(s):
        s.frames: List[Image.Image] = []

    def add(s, path):
        s.frames.append(Image.open(path).copy())

    def save(s, id, fps=10):
        s.frames[0].save(
            f"{id}.gif",
            format="GIF",
            append_images=s.frames[1:],
            save_all=True,
            duration=int(1000 / fps),
            loop=0,
            optimize=True,
        )
